/*
 * WSL (Web Scripting Language) grammar, Lex specification
 *
 * $author Pietro Spinella (pietro.spinella@gmail.com)
 * $author Eugenio Siciliano (geggio.sunrise@gmail.com)
 */

	// Set the current position (line and column, for error handling)
	void setCurrentPosition();
	// Remove the current multi-line comment (returns 0 if EOF is found in comment, 1 instead)
	int removeCurrentMultiLineComment();
	// Purge the given quoted string removing quotes and escape sequences; the string returned MUST be freed.
	char *purgeString(char *);

	// Store the current line and column
	int currentLine = 1;
	int currentColumn = 1;

literal		[[:upper:][:lower:]]
digit		[[:digit:]]
identifier	({literal}|_)({literal}|{digit}|_)*
quotedString	\"([^"\\]|\\\\|\\\")*\"

%%

"//".*$		{ setCurrentPosition(); }
"/*"		{ setCurrentPosition();
			if(!removeCurrentMultiLineComment())
			{
				yyterminate();
				return;
			}
		}

SERVICE		{ setCurrentPosition(); return SERVICE; }
INPUT		{ setCurrentPosition(); return INPUT; }
PROCEDURE	{ setCurrentPosition(); return PROCEDURE; }
CYCLE		{ setCurrentPosition(); return CYCLE; }
MATCHES		{ setCurrentPosition(); return MATCHES; }
REQUEST		{ setCurrentPosition(); return REQUEST; }
CONTINUE	{ setCurrentPosition(); return CONTINUE; }
BREAK		{ setCurrentPosition(); return BREAK; }
RETURN		{ setCurrentPosition(); return RETURN; }

id		{ setCurrentPosition(); return ID; }
content		{ setCurrentPosition(); return CONTENT; }
type		{ setCurrentPosition(); return TYPE; }
regex		{ setCurrentPosition(); return REGEX; }
name		{ setCurrentPosition(); return NAME; }
icon		{ setCurrentPosition(); return ICON; }
description	{ setCurrentPosition(); return DESCRIPTION; }
defaultValue	{ setCurrentPosition(); return DEFAULT_VALUE; }
version		{ setCurrentPosition(); return VERSION; }
author		{ setCurrentPosition(); return AUTHOR; }
enabled		{ setCurrentPosition(); return ENABLED; }

onTrue		{ setCurrentPosition(); return ON_TRUE; }
onFalse		{ setCurrentPosition(); return ON_FALSE; }
onTrueOutput	{ setCurrentPosition(); return ON_TRUE_OUTPUT; }
onFalseOutput	{ setCurrentPosition(); return ON_FALSE_OUTPUT; }

"="		{ setCurrentPosition(); return ASSIGNMENT_OP; }
"+"		{ setCurrentPosition(); return PLUS_OP; }
"-"		{ setCurrentPosition(); return MINUS_OP; }
"*"		{ setCurrentPosition(); return MULTIPLICATION_OP; }
"/"		{ setCurrentPosition(); return DIVIDE_OP; }
"%"		{ setCurrentPosition(); return PERCENT_OP; }

{identifier}	{ setCurrentPosition();
			yylval.stype = strdup(yytext);
			// TODO: free(yylval.stype);
			return IDENTIFIER;
		}
{digit}*	{ setCurrentPosition();
			yylval.stype = strdup(yytext);
			// TODO: free(yylval.stype);
			return INTEGER;
		}
{quotedString}	{ setCurrentPosition();
			yylval.stype = purgeString(yytext);
			// TODO: free(yylval.stype);
			return QUOTED_STRING;
		}

(" "|\t)+	{ setCurrentPosition(); }
\n		{ setCurrentPosition(); return NEWLINE; }

.		{ setCurrentPosition(); printf("[wsc::scanner] WARNING: character '%s' ignored at line %d, column %d\n", yytext, currentLine, currentColumn); }

%%

// Set the current position (line and column, for error handling)
void setCurrentPosition()
{
	register int i;

	for(i = 0; yytext[i] != '\0'; i++)
	{
		if(yytext[i] != '\n')
		{
			if(yytext[i] != '\t')
			{
				currentColumn++;
			} else {
				currentColumn += 8 - (currentColumn % 8);
			}
		} else {
			currentColumn = 0;
			currentLine++;
		}
	}
}

// Remove the current multi-line comment (returns 0 if EOF is found in comment, 1 instead)
int removeCurrentMultiLineComment()
{
	register int c;
	int result = 1;

	while(1)
	{
		while((c = input()) != '*' && c != EOF)
		{
			if(c != '\n')
			{
				currentColumn++;
			} else {
				currentColumn = 0;
				currentLine++;
			}
		}
		if(c == '*')
		{
			while((c = input()) == '*' && c != EOF)
			{
				currentColumn++;
			}
			if(c == '\n')
			{
				currentColumn = 0;
				currentLine++;
			}
			if(c != EOF) {
				currentColumn++;
				if(c == '/')
				{
					break;
				}
			}
		}
		if(c == EOF) {
			fprintf(stderr, "[wsc::scanner] ERROR: end of file (EOF) in comment\n");
			result = 0;
			break;
		}
	}

	return result;
}

// Purge the given quoted string removing quotes and escape sequences; the string returned MUST be freed.
char *purgeString(char *s)
{
	int len = strlen(s);
	char *r = (char *) malloc(len - 2 + 1);
	int i, j;

	for(i = 1, j = 0; i < len - 1; i++, j++)
	{
		char c = s[i];
		if(c != '\\')
		{
			r[j] = c;
		} else {
			r[j] = s[++i];
		}
	}
	r[j] = '\0';

	return r;
}
