/*
 * WSL (Web Scripting Language) grammar, Bison specification
 *
 * $author Eugenio Siciliano (geggio.sunrise@gmail.com)
 * $author Pietro Spinella (pietro.spinella@gmail.com)
 */

%{
	#include <stdio.h>
	#include <string.h>

	#include "hashtable.h"
	#include "hashtable_itr.h"
	#include "requestsplitter.h"

	// Called by the parser on syntax/parse error
	void yyerror(char const *);

	// Entry point
	int main(int argc, char *argv[]);

	// Basic functions for hashtables' handling
	static unsigned int hashfromkey(void *ky);
	static int equalkeys(void *k1, void *k2);

	// Allocate and deallocate memory
	void initialize();
	void finish();

	// Generate the xml
	void generateXml(FILE *output);

	// Stores the hashtables
	struct hashtable *identifierTable;
	struct hashtable *inputTable;
	struct hashtable *outputTable;
	struct hashtable *actionTable;
	struct hashtable *sequenceTable;

	// Key type (for all the hashtables)
	typedef struct key {
		char *str;
	} key_type;

	// Identifier type (for identifiers' hashtable)
	typedef struct _identifier {
		char* id;	// Name
		char* scope;	// Scope
		int isInput;	// True if it's an input
	} identifier_type;

	// Service type
	typedef struct {
		char* id;
		char* name;
		char* icon;
		char* description;
		char* version;
		char* author;
		char* enabled;
	} service_type;

	// Input type
	typedef struct {
		char *id;
		char *content;
		char *type;
		char *regex;
		char *name;
		char *description;
		char *defaultValue;
	} input_type;

	// Output type
	typedef struct {
		char *id;
		char *content;
		char *source;
	} output_type;

	// Action type
	typedef struct {
		int isGET;
		int isPOST;
		char *id;		// unique name
		char *iid;		// internal name
		char *containsRegex;
		char *onTrueOutput;
		char *onFalseOutput;
		char *onTrue;
		char *onFalse;
		char *content;
	} action_type;

	// Input node type (IT: questo tipo rappresenta un nodo di una (ed una) lista di input, l'id serve a trovare l'input nella corrispettiva hashtable)
	typedef struct _inputNode {
		struct _inputNode* nextInput;
		char* id;
	} input_node_type;
	
	// Sequence node type (IT: Questo tipo rappresenta un nodo di una (ed una) lista di sequence, l'id serve a trovare la sequence nella corrispettiva hashtable)
	typedef struct _sequenceNode {
	    struct _sequenceNode* nextSequence;
	    input_node_type* inputlist;
	    char* id;
	} sequence_node_type;

	// Stores the current service, input, action and sequence list
	service_type* currentService;
	input_type* currentInput;
	action_type* currentAction;
	sequence_node_type* currentSequence;

	// Store some counts
	int inputs_count = 0;
	int outputs_count = 0;
	int actions_count = 0;
	int sequences_count = 0;

	// Stores a buffer used by sprintf
	char str[20];

	// Macros for identifiers' hashtable
	DEFINE_HASHTABLE_INSERT(insert_identifier, key_type, identifier_type);
	DEFINE_HASHTABLE_SEARCH(search_identifier, key_type, identifier_type);
	DEFINE_HASHTABLE_REMOVE(remove_identifier, key_type, identifier_type);
	//DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_identifier, key_type);
	
	// Macros for inputs' hashtable
	DEFINE_HASHTABLE_INSERT(insert_input, key_type, input_type);
	DEFINE_HASHTABLE_SEARCH(search_input, key_type, input_type);
	DEFINE_HASHTABLE_REMOVE(remove_input, key_type, input_type);
	//DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_input, key_type);
	
	// Macros for outputs' hashtable
	DEFINE_HASHTABLE_INSERT(insert_output, key_type, output_type);
	DEFINE_HASHTABLE_SEARCH(search_output, key_type, output_type);
	DEFINE_HASHTABLE_REMOVE(remove_output, key_type, output_type);
	//DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_output, key_type);
	
	// Macros for actions' hashtable
	DEFINE_HASHTABLE_INSERT(insert_actions, key_type, action_type);
	DEFINE_HASHTABLE_SEARCH(search_actions, key_type, action_type);
	DEFINE_HASHTABLE_REMOVE(remove_actions, key_type, action_type);
	//DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_actions, key_type);
	
	// Macros for sequences' hashtable
	DEFINE_HASHTABLE_INSERT(insert_sequence, key_type, sequence_node_type);
	DEFINE_HASHTABLE_SEARCH(search_sequence, key_type, sequence_node_type);
	DEFINE_HASHTABLE_REMOVE(remove_sequence, key_type, sequence_node_type);
	//DEFINE_HASHTABLE_ITERATOR_SEARCH(search_itr_sequence, key_type);

	// TODO: Re-ingegnerize!
	// Functions invoked by the grammar
	void do_input( char* identifier );
	void do_inputContent( char* string );
	void do_inputType( char* string );
	void do_assignment( char* id, char* val );
	char* do_idPlusStuff( char* first, char* second );
	char* do_identifier( char* id );
	void do_assignReqGET( char* id, char* url );
	void do_assignReqPOST( char* id, char* url, char* args );
	char* do_stringPlusText( char* first, char* second );
	char* newRequestId();
	void do_procedure( char* name );
	void addActionToSequence( char* actionName );
	void addInputToSequence( char* inputName );
	char* do_show( char* text );
	void do_check( char* id, char* regex );

%}

%token SERVICE INPUT PROCEDURE CYCLE MATCHES REQUEST CONTINUE BREAK RETURN
%token ID CONTENT TYPE REGEX NAME ICON DESCRIPTION DEFAULT_VALUE VERSION AUTHOR ENABLED
%token ON_TRUE ON_FALSE ON_TRUE_OUTPUT ON_FALSE_OUTPUT
%token ASSIGNMENT_OP PLUS_OP MINUS_OP MULTIPLICATION_OP DIVIDE_OP PERCENT_OP
%token IDENTIFIER INTEGER QUOTED_STRING
%token NEWLINE

// TODO: Remove shift/reduce conflicts
%expect 27

%union {
	char* stype;
}

%type <stype> strings string integer_expr number INTEGER QUOTED_STRING

//%right ASSIGNMENT_OP
%left PLUS_OP

%%

start			: newlines service newlines inputs_or_procedures
			;
newlines		:
			| newlines NEWLINE
			;
service			: SERVICE NEWLINE service_attributes
			;
service_attributes	:
			| service_attributes ID          ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->id = $<stype>4; }
			| service_attributes NAME        ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->name = $<stype>4; }
			| service_attributes ICON        ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->icon = $<stype>4; }
			| service_attributes DESCRIPTION ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->description = $<stype>4; }
			| service_attributes VERSION     ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->version = $<stype>4; }
			| service_attributes AUTHOR      ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->author = $<stype>4; }
			| service_attributes ENABLED     ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentService->enabled = $<stype>4; }
			;
inputs_or_procedures	:
			| inputs_or_procedures input newlines
			| inputs_or_procedures procedure newlines
			;
input			: input_head input_attributes newlines
			;
input_head		: INPUT	IDENTIFIER NEWLINE							{ do_input($<stype>2); }
			;
input_attributes	:
			| input_attributes CONTENT       ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ do_inputContent( $4 ); }
			| input_attributes TYPE          ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ do_inputType( $4 ); }
			| input_attributes REGEX         ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentInput->regex = $4; }
			| input_attributes NAME          ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentInput->name = $4; }
			| input_attributes DESCRIPTION   ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentInput->description = $4; }
			| input_attributes DEFAULT_VALUE ASSIGNMENT_OP QUOTED_STRING NEWLINE		{ currentInput->defaultValue = $4; }
			;
procedure		: procedure_head blocks RETURN NEWLINE
			;
procedure_head		: PROCEDURE IDENTIFIER newlines							{ do_procedure($<stype>2); }
			;
blocks			: newlines
			| blocks assignment newlines
			| blocks command newlines
			| blocks matches newlines
			| blocks cycle newlines
			;
assignment		: IDENTIFIER ASSIGNMENT_OP assignment_value NEWLINE				{ do_assignment( $<stype>1, $<stype>3 ); }
			;
assignment_value	: identifier									{ $<stype>$ = $<stype>1; }
			| assignment_value PLUS_OP identifier						{ $<stype>$ = do_idPlusStuff( $<stype>1, $<stype>3 ); }
			;
identifier		: IDENTIFIER									{ $<stype>$ = do_identifier( $<stype>1 ); }
			| QUOTED_STRING									{ $<stype>$ = $<stype>1; }
			| INTEGER									{ $<stype>$ = $<stype>1; }
			;
command			: REQUEST strings NEWLINE							{ do_assignReqGET( newRequestId(), $<stype>2 ); }
			| REQUEST strings strings NEWLINE						{ do_assignReqPOST( newRequestId(), $<stype>2, $<stype>3 ); }
			| IDENTIFIER ASSIGNMENT_OP REQUEST strings NEWLINE				{ do_assignReqGET( $<stype>1, $<stype>4 ); }
			| IDENTIFIER ASSIGNMENT_OP REQUEST strings strings NEWLINE			{ do_assignReqPOST( $<stype>1, $<stype>4, $<stype>5 ); }
			;
matches			: matches_head matches_attributes
			;
matches_head		: MATCHES IDENTIFIER strings NEWLINE						{ do_check( $<stype>2, $<stype>3 ); }
			;
matches_attributes	:
			| matches_attributes ON_TRUE         ASSIGNMENT_OP BREAK         NEWLINE	{ currentAction->onTrue = "break"; }
			| matches_attributes ON_TRUE         ASSIGNMENT_OP CONTINUE      NEWLINE	{ currentAction->onTrue = "continue"; }
			| matches_attributes ON_FALSE        ASSIGNMENT_OP BREAK         NEWLINE	{ currentAction->onFalse = "break"; }
			| matches_attributes ON_FALSE        ASSIGNMENT_OP CONTINUE      NEWLINE	{ currentAction->onFalse = "continue"; }
			| matches_attributes ON_TRUE_OUTPUT  ASSIGNMENT_OP QUOTED_STRING NEWLINE	{ currentAction->onTrueOutput = do_show( $<stype>3 ); }
			| matches_attributes ON_FALSE_OUTPUT ASSIGNMENT_OP QUOTED_STRING NEWLINE	{ currentAction->onFalseOutput = do_show( $<stype>3 ); }
			;
cycle			: cycle_head newlines blocks BREAK NEWLINE
			;
cycle_head		: CYCLE integer_expr NEWLINE
			| CYCLE	NEWLINE
			;
strings			: string									{ $<stype>$ = $<stype>1; }
			| strings PLUS_OP string							{ $<stype>$ = do_stringPlusText( $<stype>1, $<stype>3 ); }
			;
string			: QUOTED_STRING						
			| IDENTIFIER									{ $<stype>$ = do_identifier( $<stype>1 ); }
			;
integer_expr		: number									{ $$ = $1; }
			| integer_expr PLUS_OP integer_expr						{ sprintf(str, "%d", atoi($1) + atoi($3) ); $$ = str; }
			| integer_expr MINUS_OP integer_expr						{ sprintf(str, "%d", atoi($1) - atoi($3) ); $$ = str; }
			| integer_expr MULTIPLICATION_OP integer_expr					{ sprintf(str, "%d", atoi($1) * atoi($3) ); $$ = str; }
			| integer_expr DIVIDE_OP integer_expr						{ sprintf(str, "%d", atoi($1) / atoi($3) ); $$ = str; }
			| integer_expr PERCENT_OP integer_expr						{ sprintf(str, "%d", atoi($1) % atoi($3) ); $$ = str; }
			;
number			: INTEGER									{ $<stype>$ = $<stype>1; }
			| IDENTIFIER									{ $<stype>$ = $<stype>1; }
			;

%%

#include "lex.yy.c"

// Called by the parser on syntax or parse error
void yyerror(char const *s)
{
	fprintf(stderr, "Line %d: %s\n", currentLine, s);
}

// Entry point
int main(int argc, char *argv[])
{
	// Stores the result of the yyparse function
	int result = 0;

	// Stores the output file descriptor (default = stdout)
	FILE *output = stdout;

	// Set the input and output files according to the given arguments
	switch(argc)
	{
		case 2:
			if((yyin = fopen(argv[1], "r")) == 0)
			{
				fprintf(stderr, "wsc: %s: no such file\n", argv[1]);
				return 0;
			} else {
				int len = strlen(argv[1]);
				char *str = (char *) malloc(len + 4 + 1);
				strcpy(str, argv[1]);
				if(len >= 4 && strcmp(&str[len - 4], ".wsl") == 0)
				{
					len -= 4;
				}
				str[len + 0] = '.';
				str[len + 1] = 'x';
				str[len + 2] = 'm';
				str[len + 3] = 'l';
				str[len + 4] = '\0';
				if((output = fopen(str, "w")) == 0)
				{
					fclose(yyin);
					fprintf(stderr, "wsc: %s: can't open file in write mode\n", str);
					free(str);
					return 0;
				}
				free(str);
			}
			break;
		case 3:
			if((yyin = fopen(argv[1], "r")) == 0)
			{
				fprintf(stderr, "wsc: %s: no such file\n", argv[1]);
				return 0;
			} else {
				if((output = fopen(argv[2], "w")) == 0)
				{
					fprintf(stderr, "wsc: %s: can't open file in write mode\n", argv[2]);
					fclose(yyin);
					return 0;
				}
			}
			break;
		default:
			fprintf(stderr, "Usage: wsc <source_file> [<destination_file>]\n");
			return 0;
	}

	// Set the internal structures, invoke the parser and write the output
	if((identifierTable = create_hashtable(8, hashfromkey, equalkeys)) != 0) {
		if((inputTable = create_hashtable(8, hashfromkey, equalkeys)) != 0) {
			if((outputTable = create_hashtable(8, hashfromkey, equalkeys)) != 0) {
				if((actionTable = create_hashtable(8, hashfromkey, equalkeys)) != 0) {
					if((sequenceTable = create_hashtable(8, hashfromkey, equalkeys)) != 0) {
						initialize();
						if((result = yyparse()) == 0) {
							generateXml(output);
						}
						finish();
						hashtable_destroy(actionTable, 1);
					}
					hashtable_destroy(sequenceTable, 1);
				}
				hashtable_destroy(inputTable, 1);
			}
			hashtable_destroy(outputTable, 1);
		}
		hashtable_destroy(identifierTable, 1);
	}

	// Close the input and output files
	fclose(yyin);
	fclose(output);

	// Return the result
	return result;

}

// See hashtable.h
static unsigned int hashfromkey(void *ky)
{
	char *str = ((key_type *) ky)->str;
	unsigned long hash = 0;
	int c;

	while(c = *str++) {
		hash = c + (hash << 6) + (hash << 16) - hash;
	}

	return (unsigned int) hash;

}

// See hashtable.h
static int equalkeys(void *k1, void *k2)
{
	return !strcmp(((key_type *) k1)->str, ((key_type *) k2)->str);
}

// Allocate memory
void initialize()
{
	currentService = (service_type *) malloc(sizeof(service_type));
	if(currentService != NULL)
	{
		currentService->id = NULL;
		currentService->name = NULL;
		currentService->icon = NULL;
		currentService->description = NULL;
		currentService->version = NULL;
		currentService->author = NULL;
		currentService->enabled = NULL;
	}
}

// Deallocate memory
void finish()
{
	if(currentService != NULL)
	{
		if(currentService->id != NULL) free(currentService->id);
		if(currentService->name != NULL) free(currentService->name);
		if(currentService->icon != NULL) free(currentService->icon);
		if(currentService->description != NULL) free(currentService->description);
		if(currentService->version != NULL) free(currentService->version);
		if(currentService->author != NULL) free(currentService->author);
		if(currentService->enabled != NULL) free(currentService->enabled);
		free(currentService);
	}
}

// Generate the xml
void generateXml(FILE *output)
{
	// <kmsRoot>
	fprintf(output, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	fprintf(output, "<kmsRoot version=\"0.1\">\n");

		// <services>
		fprintf(output, "\t<services>\n");

			// <service>
			fprintf(output, "\t\t<service id=\"%s\" name=\"%s\" icon=\"%s\" description=\"%s\" version=\"%s\" author=\"%s\" enabled=\"%s\">\n", currentService->id != NULL ? currentService->id : "", currentService->name != NULL ? currentService->name : "", currentService->icon != NULL ? currentService->icon : "", currentService->description != NULL ? currentService->description : "", currentService->version != NULL ? currentService->version : "", currentService->author != NULL ? currentService->author : "", currentService->enabled != NULL ? currentService->enabled : "");

			// Inputs
			if(hashtable_count(inputTable) > 0)
			{
				fprintf(output, "\t\t\t<inputs>\n");
				struct hashtable_itr *itr = hashtable_iterator(inputTable);
				do {
					input_type *v = (input_type *) hashtable_iterator_value(itr);
					fprintf(output, "\t\t\t\t<input id=\"%s\" content=\"%s\" type=\"%s\" regex=\"%s\" name=\"%s\" description=\"%s\" defaultValue=\"%s\"/>\n", v->id, v->content, v->type, v->regex, v->name, v->description, v->defaultValue);
				} while(hashtable_iterator_advance(itr));
				fprintf(output, "\t\t\t</inputs>\n");
			}

			// Outputs
			if(hashtable_count(outputTable) > 0)
			{
				fprintf(output, "\t\t\t<outputs>\n");
				struct hashtable_itr *itr = hashtable_iterator(outputTable);
				do {
					output_type *v = (output_type *) hashtable_iterator_value(itr);
					fprintf(output, "\t\t\t\t<output id=\"%s\" content=\"%s\" source=\"%s\"/>\n", v->id, v->content, v->source);
				} while(hashtable_iterator_advance(itr));
				fprintf(output, "\t\t\t</outputs>\n");
			}

			// Actions
			if(hashtable_count(actionTable) > 0)
			{
				fprintf(output, "\t\t\t<actions>\n");
				struct hashtable_itr *itr = hashtable_iterator(actionTable);
				int id = 0;
				do {
					action_type *v = (action_type *) hashtable_iterator_value(itr);
					if(v->isGET==1 || v->isPOST==1)
					{
						fprintf(output, "\t\t\t\t<action id=\"%s\" type=\"%s\">\n", v->id, "httpRequest");
						fprintf(output, "\t\t\t\t\t<httpRequest id=\"%s\">\n", v->iid );
						// Span string (BEGIN)
						char *str = v->content;
						int len = strlen(v->content);
						int offset = 0;
						int index;
						for(index = 0; ; index++)
						{
							for(; index < len && str[index] != '\n'; index++);
							if(index > len)
							{
								break;
							} else {
								str[index] = '\0';
								fprintf(output, "\t\t\t\t\t\t%s\n", &str[offset]);
								str[index] = '\n';
								offset = index + 1;
							}
						}
						// Span string (END)
						fprintf(output, "\t\t\t\t\t</httpRequest>\n");
						fprintf(output, "\t\t\t\t</action>\n");
						id++;
					} else {
						fprintf(output, "\t\t\t\t<action id=\"%s\" type=\"%s\">\n", v->id, "checkHttpRequest");
						fprintf(output, "\t\t\t\t\t<checkHttpRequest httpRequest=\"%s\" containsRegex=\"%s\"", v->iid, v->containsRegex);
						if(v->onTrue != NULL)
						{
							fprintf(output, " onTrue=\"%s\"", v->onTrue);
						}
						if(v->onFalse != NULL)
						{
							fprintf(output, " onFalse=\"%s\"", v->onFalse);
						}
						if(v->onTrueOutput != NULL)
						{
							fprintf(output, " onTrueOutput=\"%s\"", v->onTrueOutput);
						}
						if(v->onFalseOutput != NULL)
						{
							fprintf(output, " onFalseOutput=\"%s\"", v->onFalseOutput);
						}
						fprintf(output, "/>\n");
						fprintf(output, "\t\t\t\t</action>\n");
					}
				} while(hashtable_iterator_advance(itr));
				fprintf(output, "\t\t\t</actions>\n");
			}

			// Sequences
			if(hashtable_count(sequenceTable) > 0)
			{
				fprintf(output, "\t\t\t<sequences>\n");
				struct hashtable_itr *itr = hashtable_iterator(sequenceTable);
				do {
					sequence_node_type *v = (sequence_node_type *) hashtable_iterator_value(itr);
					sequence_node_type* currentSequence = v;
					fprintf(output, "\t\t\t\t<sequence id=\"%s\">\n", v->id);
					input_node_type *currentInput = currentSequence->inputlist;
					if(currentInput != NULL)
					{
						fprintf(output, "\t\t\t\t\t<usedInputs>\n");
						for(; currentInput != NULL; currentInput = currentInput->nextInput)
						{
							fprintf(output, "\t\t\t\t\t\t<usedInput input=\"%s\"/>\n", currentInput->id);
						}
						fprintf(output, "\t\t\t\t\t</usedInputs>\n");
					}
					sequence_node_type* currentAction = currentSequence->nextSequence;
					if(currentAction != NULL)
					{
						fprintf(output, "\t\t\t\t\t<usedActions>\n");
						for(; currentAction != NULL; currentAction = currentAction->nextSequence)
						{
							key_type *k;
							action_type *found;
							k = (key_type *) malloc(sizeof(key_type));
							k->str = currentAction->id;
							found = search_actions(actionTable, k);
							free(k);
							if(found != NULL)
							{
								fprintf(output, "\t\t\t\t\t\t<usedAction action=\"%s\"/>\n", found->id);
							} else {
								fprintf(stderr, "Warning: action [%s] not defined\n", currentAction->id);
							}
						}
						fprintf(output, "\t\t\t\t\t</usedActions>\n");
					}
					fprintf(output, "\t\t\t\t</sequence>\n");
				}
				while(hashtable_iterator_advance(itr));
				fprintf(output, "\t\t\t</sequences>\n");
			}

			// </service>
			fprintf(output, "\t\t</service>\n");

		// </services>
		fprintf(output, "\t</services>\n");

	// </kmsRoot>
	fprintf(output, "</kmsRoot>\n");

}

void do_input( char* identifier ) {
	
	inputs_count++;
	key_type *k, *kk;
	identifier_type *v;
	char tmp[8+strlen( identifier)];
	int i;
	
	k = (key_type *)malloc(sizeof(key_type));
	kk = (key_type *)malloc(sizeof(key_type));
	k->str = identifier;
	kk->str = identifier;
	
	//int size = 
	v = (identifier_type *)malloc(sizeof(identifier_type));
	
/*	tmp[0] = '$';
	tmp[1] = 'I';
	tmp[2] = 'N';
	tmp[3] = 'P';
	tmp[4] = 'U';
	tmp[5] = 'T';
	tmp[6] = '[';
	tmp[7] = '\0';
	
	strncat( tmp, k->str, 100 );
	v->id = strdup( strcat( tmp, "]" ) );*/	// TOLTA DA SCROCKMAN
	v->id = strdup( k->str );	// AGGIUNTA DA SCROCKMAN
	v->scope = "";
	v->isInput = 1;
	
	//printf( "Resulting string: \"%s\" for [%s]\n", v->id, k->str );
	
	if (!insert_identifier(identifierTable,k,v)){
		printf( "Unrecoverable error\n" );
		exit(-1);
	}

	//-------------------- generating input_type
	input_type* inOb = (input_type *)malloc(sizeof(input_type));
	inOb->id = identifier;
	inOb->content = "";
	inOb->type = "";
	inOb->regex = "";
	inOb->name = "";
	inOb->description = "";
	inOb->defaultValue = "";
	
	if (!insert_input(inputTable,kk,inOb)){
		printf( "Unrecoverable error\n" );
		exit(-1);
	}
	
	currentInput = inOb;
	
}

void do_inputContent( char* string ){
	// checking keywords
	if( strcmp(string,"string") &&
		strcmp(string,"password") &&
		strcmp(string,"text") &&
		strcmp(string,"collection") &&
		strcmp(string,"data") ){
		printf( "\nWarning: value \"%s\" not recognized", string );
	}
	
	currentInput->content = string;
}

void do_inputType( char* string ){
	// checking keywords
	if( strcmp(string,"constant") &&
		strcmp(string,"variable") &&
		strcmp(string,"contextual") ){
		printf( "\nWarning: value \"%s\" not recognized", string );
	}
	currentInput->type = string;
}

void do_assignment( char* id, char* val ){
	key_type *k;
	identifier_type *v;
	//printf( "aaa\n" );
	k = (key_type *)malloc(sizeof(key_type));
	k->str = strdup(id);
	
	v = (identifier_type *)malloc(sizeof(identifier_type));
	v->id = strdup(val);
	
	v->scope = currentSequence->id;
	v->isInput = 0;
	
	if (!insert_identifier(identifierTable,k,v)) exit(-1);
	
	//printf( "Assignment: [%s] = \"%s\"\n", id, val );
	
	//printf( "Inserted value \"%s\" in [%s]\n", v->id, k->str );
	//printf( "Assigned %s\n", id );
}

char* do_idPlusStuff( char* first, char* second ){
	char* result;
    
    //printf( "First is %s and Second is %s\n", first, second );
    
	if( atoi(first) && atoi(second) ){
		// sum
		sprintf(str, "%d", atoi(first) + atoi(second) );
		result = str;
		//printf( "[%s] + [%s] = [%s]\n", first, second, result );
	}
	else{
		int size = strlen(first) + strlen(second) + 1;
		result = (char*)malloc( sizeof(char)*size );
		result[0] = '\0';
		// cat
		result = strcat( result, first );
		result = strcat( result, second );
		//printf( "cat: %s\n", result );
	}
	return result;
}

char* do_identifier( char* id ){
	char* result;
	key_type *k;
    identifier_type *found;
	
	//printf( "bbbbbbb\n" );
	
	//key_type *kk;
    identifier_type *v;
	//struct hashtable_itr *itr;
    int i;
	
	k = (key_type *)malloc( sizeof(key_type) );
	k->str = id;
	
	//printf( "Searching for identifier [%s]...\n", k->str );
	//printf( "(hashtable contains %u items)\n", hashtable_count(identifierTable) );
	
	/*	
		itr = hashtable_iterator(identifierTable);
	 i = 0;
	 if (hashtable_count(identifierTable) > 0){
		 do {
			 kk = (key_type*)hashtable_iterator_key(itr);
			 v = (identifier_type*)hashtable_iterator_value(itr);
			 printf( "[%s] is \"%s\"\n", kk->str, v->id );
			 i++;
			 
		 } while (hashtable_iterator_advance(itr));
	 }
	 printf("- Iterated through %u entries -\n", i);*/
	
	found = search_identifier(identifierTable,k);
	//found = hashtable_search( identifierTable, (void*)k );
	if( NULL == found ){
		printf( "Warning: identifier [%s] not assigned\n", id );
		result = "<null>";
	}
	else{
		
		if( found->isInput ){
			char tmp[8+strlen(found->id)];
			tmp[0] = '$';
			tmp[1] = 'I';
			tmp[2] = 'N';
			tmp[3] = 'P';
			tmp[4] = 'U';
			tmp[5] = 'T';
			tmp[6] = '[';
			tmp[7] = '\0';
	
			strncat( tmp, found->id, 100 );
			result = strdup( strcat( tmp, "]" ) );	// TOLTA DA SCROCKMAN

			// add this input to the used inputs' list
			addInputToSequence( found->id );
		}
		else{
			//printf( "[%s] is \"%s\"\n", id, found->id );
			result = strdup( found->id );
			//result = found->id;
		}
	}
	return result;
}

void do_assignReqGET( char* id, char* url ){
	//printf( "The id is: \"%s\"\n", id );
	actions_count++;
	
	key_type *k;
	action_type *v;
	
	k = (key_type *)malloc(sizeof(key_type));
	k->str = id;
	
	v = (action_type *)malloc(sizeof(action_type));
	v->isGET  = 1;
	v->isPOST = 0;
	v->id = id;

	char* internal = (char*)malloc(sizeof(char)*(strlen(v->id)+2));
	internal[0] = '\0';
	internal = strcat( internal, strdup(v->id) );
	v->iid = strcat( internal, "_internal" );
	
	char* get = "GET ";
	char* http = " HTTP/1.1\\r\\n\n";
	char* host = "Host: ";
	char* ret = "\\r\\n";
	
	int size =	strlen( url ) +
		strlen(get) + strlen(http) +
		strlen(host) +
		strlen(ret) + 1;
	char* temp = (char*)malloc(sizeof(char)*size);
	
	temp[0] = '\0';
	temp = strcat( temp, get );
	//printf( "The url is: \"%s\"\n", url );
	//printf( "The string is: \"%s\"\n", extractPageFromUri(url) );
	//char* mystr = strcpy(mystr,extractPageFromUri(url));
	temp = strcat( temp, extractPageFromUri(url) );
	temp = strcat( temp, http );
	temp = strcat( temp, host );
	temp = strcat( temp, extractHostFromUri(url) );
	temp = strcat( temp, ret );
	v->content = temp;
	
	if (!insert_actions(actionTable,k,v)) exit(-1);
	
	addActionToSequence( v->id );
}

void do_assignReqPOST( char* id, char* url, char* args ){
	actions_count++;
	key_type *k;
	action_type *v;
	
	k = (key_type *)malloc(sizeof(key_type));
	k->str = id;
	
	v = (action_type *)malloc(sizeof(action_type));
	v->isGET  = 0;
	v->isPOST = 1;
	v->id = id;

	char* internal = (char*)malloc(sizeof(char)*(strlen(v->id)+2));
	internal[0] = '\0';
	internal = strcat( internal, strdup(v->id) );
	v->iid = strcat( internal, "_internal" );
	
	char* post = "POST ";
	char* http = " HTTP/1.1\\r\\n\n";
	char* host = "Host: ";
	char* ctype = "Content-Type: application/x-www-form-urlencoded\\r\\n\n";
	char* accept = "Accept: text/html, */*\\r\\n\n";
	char* aenc = "Accept-Encoding: identity\\r\\n\n";
	char* agent = "User-Agent: Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; Avant Browser; Avant Browser; .NET CLR 1.1.4322; .NET CLR 2.0.50727)\\r\\n\n";
	char* connection = "Connection: keep-alive\\r\\n\n";
	char* ret = "\\r\\n\n";
	char* data = "Data: ";

	int size =	strlen(url) + strlen(args) +
		strlen(post) + strlen(http) +
		strlen(host) +
		strlen(ctype) +
		strlen(accept) +
		strlen(aenc) +
		strlen(agent) +
		strlen(connection) +
		strlen(ret) +
		strlen(data);
	
	char* temp = (char*)malloc(sizeof(char)*size);
	
	temp[0] = '\0';
	temp = strcat( temp, post );
	temp = strcat( temp, extractPageFromUri(url) );
	temp = strcat( temp, http );
	temp = strcat( temp, host );
	temp = strcat( temp, extractHostFromUri(url) );
	temp = strcat( temp, ret );
	temp = strcat( temp, ctype );
	temp = strcat( temp, accept );
	temp = strcat( temp, aenc );
	temp = strcat( temp, agent );
	temp = strcat( temp, connection );
	temp = strcat( temp, ret );
	temp = strcat( temp, data );
	temp = strcat( temp, args );
	//temp = strcat( temp, ret );
	v->content = temp;
	//printf( "WOW! Content is:\n%s\n", v->content );

	int result = insert_actions(actionTable,k,v);

	addActionToSequence( v->id );

	if (!result)
		exit(-1);
}

char* do_stringPlusText( char* first, char* second ){
	int size = strlen(first) + strlen(second) + 1;
	char* buffer = (char*)malloc( sizeof(char)*size );
	buffer[0] = '\0';
	
	buffer = strcat( buffer, first );
	buffer = strcat( buffer, second );
	
	return strdup(buffer);
}

char* newRequestId() {

	char result[50];
	key_type *k;
    action_type *found;
    action_type *v;
    int i = 1;
	
	do{
		i++;
		k = (key_type *)malloc( sizeof(key_type) );
		sprintf( result, "action_%d", i );
		k->str = result;
		
		found = search_actions(actionTable,k);
	}
	while( NULL != found );
	
	return strdup( result );
}

char* newOutputId(){
    
    char result[50];
    key_type *k;
    output_type *found;
    output_type *v;
    int i = 0;
    
    do{
	i++;
	k = (key_type *)malloc( sizeof(key_type) );
	sprintf( result, "output_%d", i );
	k->str = result;
	
	found = search_output(outputTable,k);
    }
    while( NULL != found );
    
    return strdup( result );
}

void do_procedure( char* name ){

	sequences_count++;
	
	key_type *k;
	sequence_node_type *v;
	
	k = (key_type *)malloc(sizeof(key_type));
	k->str = name;
	
	v = (sequence_node_type *)malloc(sizeof(sequence_node_type));
	v->id  = name;
	v->nextSequence = NULL;
	v->inputlist = NULL;
	
	int result = insert_sequence(sequenceTable,k,v);
	
	currentSequence = v;
	
	if (!result) exit(-1);
	
}

// todo: add this action in the related sequence
void addActionToSequence( char* actionName ){
	//int n = 0;
    sequence_node_type* currentAction = currentSequence;//->nextSequence;
	//sequence_node_type* previousAction = currentSequence;
	
	while( currentAction->nextSequence != NULL ){
		//previousAction = currentAction;
	    
		currentAction = currentAction->nextSequence;
	    //n++;
	}
	
	//printf( "Action %d: %s\n", n, currentAction->id );
	
	//printf( "Iterated %d times\n", n );
//	previousAction->nextSequence = (sequence_node_type*)malloc( sizeof(sequence_node_type) );
	
	currentAction->nextSequence = (sequence_node_type*)malloc( sizeof(sequence_node_type) );
	
	currentAction->nextSequence->id = actionName;
	currentAction->nextSequence->nextSequence = NULL;
	
	//previousAction->nextSequence->id = actionName;
	//previousAction->nextSequence->nextSequence = NULL;
}

void addInputToSequence( char* inputName ){
    sequence_node_type* currentAction = currentSequence;
    
    if( currentAction->inputlist != NULL ){
	input_node_type* currentList = currentAction->inputlist;
	
	if( !strcmp(currentList->id, inputName) ) return;
	
	while( currentList->nextInput != NULL ){
	    if( !strcmp(currentList->nextInput->id, inputName) ) return;
	    currentList = currentList->nextInput;
	    //n++;
	}
	
	//    currentAction->nextSequence = (sequence_node_type*)malloc( sizeof(sequence_node_type) );
	
	currentList->nextInput = (input_node_type*)malloc( sizeof(input_node_type) );
	
	currentList->nextInput->id = inputName;
	currentList->nextInput->nextInput = NULL;
    }
    else{
	currentAction->inputlist = (input_node_type*)malloc( sizeof(input_node_type) );
	currentAction->inputlist->id = inputName;
	currentAction->inputlist->nextInput = NULL;
    }
    
//    currentAction->nextSequence->id = actionName;
//    currentAction->nextSequence->nextSequence = NULL;
}

char* do_show( char* text ){
    outputs_count++;
    
	key_type *k;//, kk;
    output_type *vo;
    //action_type *va
    
    k = (key_type *)malloc(sizeof(key_type));
//    kk = (key_type *)malloc(sizeof(key_type));
    k->str = newOutputId();
//    kk->str = k->str;
    
    vo = (output_type *)malloc(sizeof(output_type));
    vo->id  = k->str;
    vo->content = "string";
    vo->source = text;

/*    va = (action_type *)malloc(sizeof(action_type));
    va->id  = k->str;
    va->isGET  = 0;
    va->isPOST = 0;*/
    currentAction->content = NULL;
    
    int result = insert_output(outputTable,k,vo);
    
    if (!result) exit(-1);
    
/*    result = insert_action(actionTable,kk,va);
    if (!result) exit(-1);*/
    
    return vo->id;
}

void do_check( char* id, char* regex ){
	actions_count++;
	
	key_type *k, *kk;
	action_type *found;

		action_type *va;
	
	k = (key_type *)malloc(sizeof(key_type));
	//    kk = (key_type *)malloc(sizeof(key_type));
	k->str = newRequestId();
	//    kk->str = k->str;

/*	vo = (output_type *)malloc(sizeof(output_type));
	vo->id  = k->str;
	vo->content = "string";
	vo->source = text;*/

	va = (action_type *)malloc(sizeof(action_type));
	va->id  = k->str;
//a
	kk = (key_type *) malloc(sizeof(key_type));
	kk->str = id;
	found = search_actions(actionTable, kk);
	if(found != NULL) {
		va->iid = found->iid;
	} else {
		va->iid = 0;
	}
	free(kk);
//a
//a	va->iid = id;
	va->isGET  = 0;
	va->isPOST = 0;
	va->containsRegex = regex;
	
	/*int result = insert_output(outputTable,k,vo);

	if (!result) exit(-1);*/

	int result = insert_actions(actionTable,k,va);
	if (!result) exit(-1);
	
	currentAction = va;

	addActionToSequence( va->id );

//	return va->id;
}
